@@ -54,6 +54,7 @@
 #include "v8.h"
 #include "v8threads.h"
 
+#include "platform-posix.h"
 #include "platform.h"
 #include "vm-state-inl.h"
 
@@ -91,6 +92,13 @@ void OS::SetUp() {
 }
 
 
+void OS::PostSetUp() {
+  // Math functions depend on CPU features therefore they are initialized after
+  // CPU.
+  MathSetup();
+}
+
+
 void OS::ReleaseStore(volatile AtomicWord* ptr, AtomicWord value) {
   __asm__ __volatile__("" : : : "memory");
   *ptr = value;
