@@ -51,6 +51,7 @@
 
 #include "v8.h"
 
+#include "platform-posix.h"
 #include "platform.h"
 #include "v8threads.h"
 #include "vm-state-inl.h"
@@ -107,6 +108,13 @@ void OS::SetUp() {
 }
 
 
+void OS::PostSetUp() {
+  // Math functions depend on CPU features therefore they are initialized after
+  // CPU.
+  MathSetup();
+}
+
+
 uint64_t OS::CpuFeaturesImpliedByPlatform() {
   return 0;
 }
